package java.util.concurrent.atomic;

import sun.misc.Contended;
import java.util.function.DoubleBinaryOperator;
import java.util.concurrent.ThreadLocalRandom;
import java.util.function.LongBinaryOperator;
import sun.misc.Unsafe;

abstract class Striped64 extends Number
{
    static final int NCPU;
    transient volatile Striped64.Cell[] cells;
    transient volatile long base;
    transient volatile int cellsBusy;
    private static final Unsafe UNSAFE;
    private static final long BASE;
    private static final long CELLSBUSY;
    private static final long PROBE;
    
    Striped64() {
    }
    
    final boolean casBase(long expected, long x) {
        return Striped64.UNSAFE.compareAndSwapLong(this, Striped64.BASE, expected, x);
    }
    
    final boolean casCellsBusy() {
        return Striped64.UNSAFE.compareAndSwapInt(this, Striped64.CELLSBUSY, 0, 1);
    }
    
    static final int getProbe() {
        return Striped64.UNSAFE.getInt(Thread.currentThread(), Striped64.PROBE);
    }
    
    static final int advanceProbe(int x) {
        x ^= x << 13;
        x ^= x >>> 17;
        x ^= x << 5;
        Striped64.UNSAFE.putInt(Thread.currentThread(), Striped64.PROBE, x);
        return x;
    }
    
    final void longAccumulate(long n, LongBinaryOperator longBinaryOperator, boolean b) {
        int n2;
        if ((n2 = getProbe()) == 0) {
            ThreadLocalRandom.current();
            n2 = getProbe();
            b = true;
        }
        int n3 = 0;
        while (true) {
            Striped64.Cell[] cells;
            int length;
            if ((cells = this.cells) != null && (length = cells.length) > 0) {
                Striped64.Cell cell;
                if ((cell = cells[length - 1 & n2]) == null) {
                    if (this.cellsBusy == 0) {
                        Striped64.Cell o = new Cell(n);
                        if (this.cellsBusy == 0 && this.casCellsBusy()) {
                            boolean n5 = false;
                            try {
                                Striped64.Cell[] cells4;
                                int length2;
                                int n6;
                                if ((cells4 = this.cells) != null && (length2 = cells4.length) > 0 && cells4[n6 = (length2 - 1 & n2)] == null) {
                                    cells4[n6] = o;
                                    n5 = true;
                                }
                            }
                            finally {
                                this.cellsBusy = 0;
                            }
                            if (n5 == 0) {
                                continue;
                            }
                            break;
                        }
                    }
                    n3 = 0;
                }
                else if (!b) {
                    b = true;
                }
                else {
                    long value;
                    if (cell.cas(value = cell.value, (longBinaryOperator == null) ? (value + n) : longBinaryOperator.applyAsLong(value, n))) {
                        break;
                    }
                    if (length >= Striped64.NCPU || this.cells != cells) {
                        n3 = 0;
                    }
                    else if (n3 == 0) {
                        n3 = 1;
                    }
                    else if (this.cellsBusy == 0 && this.casCellsBusy()) {
                        try {
                            if (this.cells == cells) {
                                Striped64.Cell[] cells3 = new Striped64.Cell[length << 1];
                                for (int i = 0; i < length; ++i) {
                                    cells3[i] = cells[i];
                                }
                                this.cells = cells3;
                            }
                        }
                        finally {
                            this.cellsBusy = 0;
                        }
                        n3 = 0;
                        continue;
                    }
                }
                n2 = advanceProbe(n2);
            }
            else if (this.cellsBusy == 0 && this.cells == cells && this.casCellsBusy()) {
                boolean n4 = false;
                try {
                    if (this.cells == cells) {
                        Striped64.Cell[] cells2 = new Striped64.Cell[2];
                        cells2[n2 & 0x1] = new Cell(n);
                        this.cells = cells2;
                        n4 = true;
                    }
                }
                finally {
                    this.cellsBusy = 0;
                }
                if (!(n4 == 0)) {
                    break;
                }
                continue;
            }
            else {
                long base;
                if (this.casBase(base = this.base, (longBinaryOperator == null) ? (base + n) : longBinaryOperator.applyAsLong(base, n))) {
                    break;
                }
                continue;
            }
        }
    }
    
    final void doubleAccumulate(double n, DoubleBinaryOperator doubleBinaryOperator, boolean b) {
        int n2;
        if ((n2 = getProbe()) == 0) {
            ThreadLocalRandom.current();
            n2 = getProbe();
            b = true;
        }
        int n3 = 0;
        while (true) {
            Striped64.Cell[] cells;
            int length;
            if ((cells = this.cells) != null && (length = cells.length) > 0) {
                Striped64.Cell cell;
                if ((cell = cells[length - 1 & n2]) == null) {
                    if (this.cellsBusy == 0) {
                        Striped64.Cell o = new Striped64.Cell(Double.doubleToRawLongBits(n));
                        if (this.cellsBusy == 0 && this.casCellsBusy()) {
                            boolean n5 = false;
                            try {
                                Striped64.Cell[] cells4;
                                int length2;
                                int n6;
                                if ((cells4 = this.cells) != null && (length2 = cells4.length) > 0 && cells4[n6 = (length2 - 1 & n2)] == null) {
                                    cells4[n6] = o;
                                    n5 = true;
                                }
                            }
                            finally {
                                this.cellsBusy = 0;
                            }
                            if (n5 == 0) {
                                continue;
                            }
                            break;
                        }
                    }
                    n3 = 0;
                }
                else if (!b) {
                    b = true;
                }
                else {
                    long value;
                    if (cell.cas(value = cell.value, (doubleBinaryOperator == null) ? Double.doubleToRawLongBits(Double.longBitsToDouble(value) + n) : Double.doubleToRawLongBits(doubleBinaryOperator.applyAsDouble(Double.longBitsToDouble(value), n)))) {
                        break;
                    }
                    if (length >= Striped64.NCPU || this.cells != cells) {
                        n3 = 0;
                    }
                    else if (n3 == 0) {
                        n3 = 1;
                    }
                    else if (this.cellsBusy == 0 && this.casCellsBusy()) {
                        try {
                            if (this.cells == cells) {
                                Striped64.Cell[] cells3 = new Striped64.Cell[length << 1];
                                for (int i = 0; i < length; ++i) {
                                    cells3[i] = cells[i];
                                }
                                this.cells = cells3;
                            }
                        }
                        finally {
                            this.cellsBusy = 0;
                        }
                        n3 = 0;
                        continue;
                    }
                }
                n2 = advanceProbe(n2);
            }
            else if (this.cellsBusy == 0 && this.cells == cells && this.casCellsBusy()) {
                boolean n4 = false;
                try {
                    if (this.cells == cells) {
                        Striped64.Cell[] cells2 = new Striped64.Cell[2];
                        cells2[n2 & 0x1] = new Striped64.Cell(Double.doubleToRawLongBits(n));
                        this.cells = cells2;
                        n4 = true;
                    }
                }
                finally {
                    this.cellsBusy = 0;
                }
                if (!(n4 == 0)) {
                    break;
                }
                continue;
            }
            else {
                long base;
                if (this.casBase(base = this.base, (doubleBinaryOperator == null) ? Double.doubleToRawLongBits(Double.longBitsToDouble(base) + n) : Double.doubleToRawLongBits(doubleBinaryOperator.applyAsDouble(Double.longBitsToDouble(base), n)))) {
                    break;
                }
                continue;
            }
        }
    }
    
    static {
        NCPU = Runtime.getRuntime().availableProcessors();
        try {
            UNSAFE = Unsafe.getUnsafe();
            Class<Object> clazz = Striped64.class;
            BASE = Striped64.UNSAFE.objectFieldOffset(clazz.getDeclaredField("base"));
            CELLSBUSY = Striped64.UNSAFE.objectFieldOffset(clazz.getDeclaredField("cellsBusy"));
            PROBE = Striped64.UNSAFE.objectFieldOffset(Thread.class.getDeclaredField("threadLocalRandomProbe"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
    
    @Contended
    static final class Cell
    {
        volatile long value;
        private static final Unsafe UNSAFE;
        private static final long valueOffset;
        
        Cell(long value) {
            this.value = value;
        }
        
        final boolean cas(long expected, long x) {
            return Striped64.Cell.UNSAFE.compareAndSwapLong(this, Striped64.Cell.valueOffset, expected, x);
        }
        
        static {
            try {
                UNSAFE = Unsafe.getUnsafe();
                valueOffset = Striped64.Cell.UNSAFE.objectFieldOffset(Striped64.Cell.class.getDeclaredField("value"));
            }
            catch (Exception cause) {
                throw new Error(cause);
            }
        }
    }
}
